{% extends "layout.html" %}
{% block body %}
<h3>
    Welcome to GNS3.
</h3>
<ul>
    <li><a href="https://gns3.com">Website</a></li>
    <li><a href="http://docs.gns3.com">Documentation</a></li>
    <li><a href="/static/web-ui/bundled">WebUI - bundled</a></li>
</ul>
<p>If you are looking for uploading the IOU. You can since 1.4 upload them directly from the client see: <a href="http://docs.gns3.com/1PKfYwR78QP_Z3jqxBQ1pdy6SsqM27qhvdCvSmIizRh4">this documentation</a>.</p>
{% endblock %}
